# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

raise ImportError(
    "penaltymodel.cache was removed in penaltymodel 1.0.0 and is no longer maintained. "
    "That library can be installed with `pip install penaltymodel-cache`. "
    "This command will also install a previous version of penaltymodel. "
    "If there was some specific functionality in maxgap that you need, please "
    "open a github issue at https://github.com/dwavesystems/penaltymodel or "
    "https://github.com/dwavesystems/dwave-ocean-sdk."
    )
