from penaltymodel.core.classes import *
import penaltymodel.core.classes

import penaltymodel.core.exceptions

from penaltymodel.core.interface import *
import penaltymodel.core.interface

from penaltymodel.core.package_info import __version__
import penaltymodel.core.package_info

from dimod import Vartype, SPIN, BINARY
